//************************************************************************
//*	iCruze drivers
//*		(C) 2010 by Mark Sproul (KB2ICI)
//*		msproul -at- skychariot.com
//*		If you find this code useful, please send me an email
//*
//*		This software is provided as is. 
//*	
//*	This library is free software; you can redistribute it and/or
//*	modify it under the terms of the GNU Lesser General Public
//*	License as published by the Free Software Foundation; either
//*	version 2.1 of the License, or (at your option) any later version.
//*	
//*	This library is distributed in the hope that it will be useful,
//*	but WITHOUT ANY WARRANTY; without even the implied warranty of
//*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//*	Lesser General Public License for more details.
//*
//*		http://www.cbxdragbike.com/arduino/icruze
//************************************************************************
//*	The iCruze uses the ATtiny26
//*	it has	2K FLASH
//*			128 bytes ram
//*
//*	The LCD drivers are modified from the Arduino LCD library.
//*	The primary differences is no C++ to save space
//************************************************************************
//*	Apr 18,	2010	avrdude talking to attiny26
//*	Apr 19,	2010	Successfully programmed blink in attiny26
//*	Apr 20,	2010	Writing to LCD display
//*	Apr 21,	2010	Got I2C working, Tim H told me to reduce the speed on the master end
//*					I2C does not work reliable enough, giving up on it
//*	Apr 23,	2010	Started on IRQ based bit-banging
//*	Apr 25,	2010	IRQ bitbang working perfectly
//************************************************************************


#define	F_CPU	8000000
#define	HIGH	1
#define	LOW		0

#include	<avr/io.h>
#include	<avr/interrupt.h>

#ifdef _CODEWARRIOR_
//*	I use Codewarrior for my editor, ignore this.
	void	_delay_us(int);
	void	_delay_ms(int);
	#undef ISR
	#define ISR(vector)	void vector()
	#undef sei
	#undef cli
	void cli();
	void sei();
#else
	#include	<util/delay.h>
#endif


#define	kPIN_TXD		0
#define	kPIN_RXD		2
#define	kBaudDealay_us	104		//*	9600 baud


//#define	__DEBUG__

#define	kMaxLines	2
#define	kMaxColumns	20

//************************************************************************
//*	font definitions
enum 
{
	kFont_Normal	=	0,
	kFont_Alternate,
	kFont_last
};


typedef  unsigned char boolean;

int8_t			gLCD_currentRow;
int8_t			gLCD_currentColumn;
uint8_t			gCurrentFont;
boolean			gWrapHasOccured;
boolean			gNeedToScroll;			//*	this is so we dont always have a blank line on the bottom
char			gLCD_buffer[kMaxLines][kMaxColumns + 1];


//************************************************************************
unsigned char	digitalRead(unsigned char bitNumber);
void			digitalWrite(unsigned char bitNumber, unsigned char newState);
void			LCDdriver_Init(void);
void			LCDdriver_write(uint8_t value);
void			LCDdriver_clear();
void			LCDdriver_home();
void			LCDdriver_setCursor(unsigned char col, unsigned char row);
static void		LCD_PrintString(const char *theString);

#ifdef __DEBUG__
	void	PutStringBitBang(char *theString);
	void	PutCharBitBang(char theChar);
#endif

//************************************************************************
static void	LCD_ClearScreen(void)
{
short	ii, jj;

	//*	set the screen buffer to all spaces
	for (ii=0; ii<kMaxLines; ii++)
	{
		for (jj=0; jj<kMaxColumns; jj++)
		{
			gLCD_buffer[ii][jj]	=	0x20;
		}
	}
	LCDdriver_clear();
	gLCD_currentColumn	=	0;
	gLCD_currentRow		=	0;
	gNeedToScroll		=	false;
}

//************************************************************************
static void	LCD_ClearLine(uint8_t lineNumber)
{
short	ii;

	if ((lineNumber >=0) && (lineNumber < kMaxLines))
	{
		//*	set the line buffer to all spaces
		for (ii=0; ii<kMaxColumns; ii++)
		{
			gLCD_buffer[lineNumber][ii]	=	0x20;
		}

		LCDdriver_setCursor(0, lineNumber);

		//*	now clear the line on the actual LCD screen
		for (ii = 0; ii < kMaxColumns; ii++)
		{
			LCDdriver_write(0x20);
		}
		LCDdriver_setCursor(0, lineNumber);
	}
}


//************************************************************************
static void	LCD_BumpToNextLine(boolean clearNewLine)
{
int	ii, jj;

	//*	fill the rest of the current line with spaces
	for (jj=gLCD_currentColumn; jj<kMaxColumns; jj++)
	{
		gLCD_buffer[gLCD_currentRow][jj]	=	0x20;
	}
	gLCD_currentColumn	=	0;		//*	back to the 1st column
	gLCD_currentRow++;				//*	to the next row

	if (gLCD_currentRow >= kMaxLines)
	{
		//*	move all of the data up one line in memory buffer
		for (ii = 0; ii < (kMaxLines - 1); ii++)
		{
			//*	copy the next line into this one
			for (jj=0; jj<kMaxColumns; jj++)
			{
				gLCD_buffer[ii][jj]	=	gLCD_buffer[ii + 1][jj];
			}
		}
		//*	now send the data back to the LCD screen
		for (ii = 0; ii < (kMaxLines - 1); ii++)
		{
			LCDdriver_setCursor(0, ii);
			for (jj=0; jj<kMaxColumns; jj++)
			{
				LCDdriver_write(gLCD_buffer[ii][jj]);
			//	delay(10);
			}
		}
		gLCD_currentRow	=	kMaxLines - 1;
	}
	LCDdriver_setCursor(gLCD_currentColumn, gLCD_currentRow);

	if (clearNewLine)
	{
		LCD_ClearLine(gLCD_currentRow);
	}

}

//************************************************************************
static void	LCD_DisplayPrintableChar(char theChar)
{
	LCDdriver_write(theChar);

	if ((gLCD_currentRow < kMaxLines) && (gLCD_currentColumn < kMaxColumns))
	{
		gLCD_buffer[gLCD_currentRow][gLCD_currentColumn]	=	theChar;
	}
	
	gLCD_currentColumn++;
	if (gLCD_currentColumn >= kMaxColumns)
	{
		gNeedToScroll	=	true;
	//	LCD_BumpToNextLine(true);
		gWrapHasOccured	=	true;
	}
	else
	{
		gWrapHasOccured	=	false;
	}
}





	
//************************************************************************
static void	LCD_SetFont(short whichFont)
{
	gCurrentFont	=	whichFont;
	if ((gCurrentFont < 0) || (gCurrentFont >= kFont_last))
	{
		gCurrentFont	=	kFont_Normal;
	}

}


//************************************************************************
static void	LCD_ProcessChar(char theChar)
{

	//*	this is so all 4 lines get used, when a CR comes in, we will wait
	//*	until the next char to actually do the scrolling
	if (gNeedToScroll && (theChar != 0x0A))
	{
		LCD_BumpToNextLine(true);
		gNeedToScroll	=	false;
	}
	switch(theChar)
	{
		case 0x08:						//*	back space
			gLCD_currentColumn--;
			if (gLCD_currentColumn < 0)
			{
				gLCD_currentColumn	=	0;
			}
			LCDdriver_setCursor(gLCD_currentColumn, gLCD_currentRow);
			break;

		case 0x09:						//*	tab
			LCD_DisplayPrintableChar(0x20);
			while (gLCD_currentColumn % 4)
			{
				LCD_DisplayPrintableChar(0x20);
			}
			break;

		case 0x0A:						//*	Line feed
			break;

		case 0x0B:						//*	Vertical tab (the opposite of line feed)
			gLCD_currentRow--;
			if (gLCD_currentRow < 0)
			{
				gLCD_currentRow	=	0;
			}
			LCDdriver_setCursor(gLCD_currentColumn, gLCD_currentRow);
			break;

		case 0x0C:						//*	form feed
			LCD_ClearScreen();
			break;
		
		case 0x0D:						//*	Carriage return
			if (gLCD_currentRow == (kMaxLines - 1))
			{
				//*	we dont want to do the CR if we just forced a bump because of line overflow
				if (gWrapHasOccured == false)
				{
					gNeedToScroll	=	true;
				}
			}
			else
			{
				LCD_BumpToNextLine(true);
			}
			gWrapHasOccured	=	false;
			break;
		
		case 0x0e:						//*	^N SHIFT OUT - use alternate character set
			LCD_SetFont(kFont_Alternate);
			break;

		case 0x0f:						//*	^O SHIFT IN, resume default character set
			LCD_SetFont(kFont_Normal);
			break;

		default:					// display each character to the LCD
			if (theChar >= 0x20)
			{
				switch(gCurrentFont)
				{
					case kFont_Alternate:
						LCD_DisplayPrintableChar(theChar + 0x80);
						break;
					
					case kFont_Normal:
					default:
						LCD_DisplayPrintableChar(theChar);
						break;
				
				}
			}
			break;
	}
}

#ifdef __DEBUG__
//************************************************************************
void	LCD_PrintInt(int theNum)
{
	
	LCD_ProcessChar((theNum / 100) + 0x30);
	LCD_ProcessChar((((theNum % 100) / 10) & 0x0f) + 0x30);
	LCD_ProcessChar((theNum % 10) + 0x30);
	LCD_ProcessChar(0x20);
}

#endif

//************************************************************************
static void	LCD_PrintString(const char *theString)
{
int	ii;

	ii	=	0;
	while (theString[ii] != 0)
	{
		LCD_ProcessChar(theString[ii]);
		ii++;
	}
}



#ifdef __DEBUG__
//************************************************************************
void	PutCharBitBang(char theChar)
{
char	ii;

	digitalWrite(kPIN_TXD, LOW);	//*	start bit = low
	_delay_us(kBaudDealay_us);
	
	for (ii=0; ii<8; ii++)
	{
		digitalWrite(kPIN_TXD, theChar & 0x01);

		_delay_us(kBaudDealay_us);
		theChar	=	theChar >> 1;
	}
	digitalWrite(kPIN_TXD, HIGH);	//*	stop bit = high
	_delay_us(kBaudDealay_us);

	_delay_us(kBaudDealay_us);		//*	2 stop bits
}

//************************************************************************
void	PutStringBitBang(char *theString)
{
int	ii;

	ii	=	0;
	while (theString[ii] != 0)
	{
		PutCharBitBang(theString[ii]);
		ii++;
	}
}

#endif


#pragma mark -
#pragma mark Timer routines
//************************************************************************
//*	Theory of operation
//*		Use an interrupt to check the input pin at least 2 x the bit time
//*	
//*		The bit timer is counter/timer 0  = 8000000 / 8 = 1 us per count
//*	
//*		The interrupt timer clock is 2 mhz (8mhz / 4) (the rate may very)
//*		I am using the Compare 1 interrupt
//*		when the counter gets to the value, it generates an interrupt
//*		the interrupt resets the counter.
//*		it will generate an interrupt 6 times for each bit 
//*
//*		The interrupt routine is a state machine that watches for a low
//*		Once it finds a low, it waits for 1/3 bit time. If the bit is high
//*		during this time, it sets it back to idle and ignores it
//*
//*		It then waits ONE FULL bit time (starting from 1/3 bit time after the start of the start bit)
//*		8 times and puts the bit in at bit #7 (serial data is LSB first)
//*
//*		Once it has the 8 bits, it puts it into the buffer
//*
//*		Then it waits for the line to be HIGH and puts it back to IDLE
//*
//************************************************************************
#define	BAUDRATE	9600

#define	kCOUNTER1_TIMEOUT	((F_CPU / 4) / BAUDRATE / 6)

#define	kBIT_TIME_us		(1000000 / BAUDRATE)
//************************************************************************
void	SetupTimters()
{
	cli();
	//*	set up timer 0
	TCNT0	=	0;				//*	start it off at 0
	TCCR0	=	(1 << PSR0) | (0 << CS02) | (1 << CS01) | (0 << CS00);	// Reset the prescaler and start Timer0.
	
	//*	set up timer 1
	TCCR1A	=	(0 << COM1A1) | (0 << COM1A0) |		//	Timer/Counter Comparator A disconnected from output pin OC1A.
				(0 << COM1B1) | (0 << COM1B0);		//	Timer/Counter Comparator B disconnected from output pin OC1B.


	TCCR1B	=	(0 << CTC1) |						//	 Clear Timer/Counter on Compare Match
				(0 << CS13) | (0 << CS12) | (1 << CS11) | (1 << CS10);	//*	CK/4 
	TCNT1	=	0;
	OCR1A	=	kCOUNTER1_TIMEOUT;

	//*	enable timer 1 Compare A interrupt
	TIMSK	=	(1 << OCIE1A);
	
	sei();
}


enum 
{
	kBitBangState_IDLE	=0,
	kBitBangState_Start,
	kBitBangState_RCV,
	kBitBangState_WaitForStop,
	
	kBitBangState_Last
};

#define	kRCV_BufferSize	12

volatile int	gTimeOfLastBit;
volatile char	gBitBangState;
volatile char	gBitRcvdCount;
volatile char	gCurrentChar;
volatile char	gBuff_Head;
volatile char	gBuff_Tail;
volatile char	gRcvBuffer[kRCV_BufferSize];
//************************************************************************
void	PutCharIntoBuffer(char theChar)
{
	gRcvBuffer[gBuff_Head++]	=	theChar;
	if (gBuff_Head >= kRCV_BufferSize)
	{
		gBuff_Head	=	0;
	}
}

//************************************************************************
char	GetCharFromBuffer(void)
{
char	theChar;

	while (gBuff_Tail == gBuff_Head)
	{
		//*	do nothing
	}
	theChar	=	gRcvBuffer[gBuff_Tail++];
	
	if (gBuff_Tail >= kRCV_BufferSize)
	{
		gBuff_Tail	=	0;
	}
	return(theChar);
}

//************************************************************************
ISR(TIMER1_CMPA_vect)
{
char	currentBitState;
int		currentTimer0;
int		deltaTime;

	TCNT1			=	0;					//*	reset the counter
//	currentBitState	=	digitalRead(kPIN_RXD);
	if ((PINB & (1 << kPIN_RXD)) == 0)
	{
		currentBitState	=	LOW;
	}
	else
	{
		currentBitState	=	HIGH;
	}




	currentTimer0	=	TCNT0 & 0x0ff;		//*	get the time from counter 0
	if (currentTimer0 < gTimeOfLastBit)
	{
		currentTimer0	+=	256;			//*	adust for overflow
	}

	switch(gBitBangState)
	{
		case kBitBangState_IDLE:
			if (currentBitState == LOW)
			{
				//*	we have a start bit
				gBitBangState	=	kBitBangState_Start;
				gTimeOfLastBit	=	currentTimer0 & 0x0ff;
			}
			break;


		case kBitBangState_Start:
			if (currentBitState == LOW)	//*	is it still low
			{
				deltaTime	=	currentTimer0 - gTimeOfLastBit;
				//*	wait until we are 1/3 of the way into the bit, this will verify its a good stop bit
				//*	and put our sampling to close to the center of the bit.
				if (deltaTime >= (kBIT_TIME_us / 3))
				{
					gBitRcvdCount	=	0;
					gCurrentChar	=	0;
					gTimeOfLastBit	=	currentTimer0 & 0x0ff;
					gBitBangState	=	kBitBangState_RCV;
				}
			}
			else
			{
				//*	this gets rid of noise
				gBitBangState	=	kBitBangState_IDLE;
			}
			break;
		
		case kBitBangState_RCV:
			deltaTime	=	currentTimer0 - gTimeOfLastBit;
			if (deltaTime >= kBIT_TIME_us)
			{
				gCurrentChar	=	gCurrentChar >> 1;
				if (currentBitState)
				{
					gCurrentChar	=	gCurrentChar | 0x80;
				}
				gTimeOfLastBit	=	currentTimer0 & 0x0ff;
				gBitRcvdCount++;
				if (gBitRcvdCount >= 8)
				{
					//*	time to put the char in the buffer
					PutCharIntoBuffer(gCurrentChar & 0x7f);
					gBitBangState	=	kBitBangState_WaitForStop;
					gBitRcvdCount	=	0;
				}
			}
			break;
		
		case kBitBangState_WaitForStop:
			//*	we want to wait until the line is a 1 again
			if (currentBitState == HIGH)
			{
				gBitBangState	=	kBitBangState_IDLE;
			}
			break;
	}
}


#pragma mark -

//************************************************************************
int main(void)
{
unsigned char	theChar;

	DDRB	=	0xf8;
	digitalWrite(1, HIGH);	//*	enable pull up resistors
	digitalWrite(3, HIGH);

	LCDdriver_Init();
	
#ifdef __DEBUG__
	DDRB	=	0xf8 | (1 << kPIN_TXD);
	PutStringBitBang("debug\r\n");
#endif
	
	gBitBangState			=	kBitBangState_IDLE;
	gBuff_Head				=	0;
	gBuff_Tail				=	0;
	SetupTimters();

	gCurrentFont			=	kFont_Normal;
	gLCD_currentColumn		=	0;
	gLCD_currentRow			=	0;
	gWrapHasOccured			=	false;
	gNeedToScroll			=	false;
	
	LCD_ClearScreen();
	LCD_PrintString("Hacked iCruze\rby KB2ICI");

#ifdef __DEBUG__
	_delay_ms(1000);
	LCD_PrintInt(kCOUNTER1_TIMEOUT);
	LCD_PrintInt(kBIT_TIME_us);

	_delay_ms(1000);
	LCD_ProcessChar(0x0d);
#endif

	while (1)
	{
		theChar	=	GetCharFromBuffer();
		LCD_ProcessChar(theChar);
	}
}

#pragma mark -
#pragma mark LCD routines

//***************************************************************
//	LCD routines
#define		DATA_PORT				PORTA
#define		CONTROL_PORT			PORTB
#define		kCTRL_BIT_RS			3
#define		kCTRL_BIT_RW			4
#define		kCTRL_BIT_ENABLE		5

#define	kLCD_RegisterSelect_CMD		0
#define	kLCD_RegisterSelect_DATA	1

#define	kLCD_RegisterSelect_CMD		0
#define	kLCD_RegisterSelect_DATA	1


// commands
#define	kLCD_CLEARDISPLAY			0x01
#define	kLCD_RETURNHOME				0x02
#define	kLCD_ENTRYMODESET			0x04
#define	kLCD_DISPLAYCONTROL			0x08
#define	kLCD_CURSORSHIFT			0x10
#define	kLCD_FUNCTIONSET			0x20
#define	kLCD_SETCGRAMADDR			0x40
#define	kLCD_SETDDRAMADDR			0x80

//	flags for display entry mode
#define	kLCD_ENTRYRIGHT				0x00
#define	kLCD_ENTRYLEFT				0x02
#define	kLCD_ENTRYSHIFTINCREMENT	0x01
#define	kLCD_ENTRYSHIFTDECREMENT	0x00

// flags for display on/off control
#define	kLCD_DISPLAYON				0x04
#define	kLCD_DISPLAYOFF				0x00
#define	kLCD_CURSORON				0x02
#define	kLCD_CURSOROFF				0x00
#define	kLCD_BLINKON				0x01
#define	kLCD_BLINKOFF				0x00

// flags for display/cursor shift
#define	kLCD_DISPLAYMOVE			0x08
#define	kLCD_CURSORMOVE				0x00
#define	kLCD_MOVERIGHT				0x04
#define	kLCD_MOVELEFT				0x00

// flags for function set
#define	kLCD_8BITMODE				0x10
#define	kLCD_4BITMODE				0x00
#define	kLCD_2LINE					0x08
#define	kLCD_1LINE					0x00
#define	kLCD_5x10DOTS				0x04
#define	kLCD_5x8DOTS				0x00



#define	kLCD_SetDDRAMaddress	0x80

//***************************************************************
unsigned char	digitalRead(unsigned char bitNumber)
{
	if ((PINB & (1 << bitNumber)) == 0)
	{
		return(LOW);
	}
	else
	{
		return(HIGH);
	}
}

//***************************************************************
void	digitalWrite(unsigned char bitNumber, unsigned char newState)
{
	if (newState)
	{
		PORTB	|=	1 << bitNumber;
	}
	else
	{
		PORTB	&=	~(1 << bitNumber);
	}
}

//************************************************************************
void LCDdriver_send(uint8_t value, uint8_t mode)
{
	digitalWrite(kCTRL_BIT_RS,	mode);
	digitalWrite(kCTRL_BIT_RW,	LOW);

	DATA_PORT	=	value;
	_delay_us(25);
	
	digitalWrite(kCTRL_BIT_ENABLE, HIGH);
	digitalWrite(kCTRL_BIT_ENABLE, LOW);
}


//************************************************************************
void	LCDdriver_command(uint8_t value)
{
	LCDdriver_send(value, kLCD_RegisterSelect_CMD);
}

//************************************************************************
void LCDdriver_write(uint8_t value)
{
	LCDdriver_send(value, kLCD_RegisterSelect_DATA);
}

//************************************************************************
void LCDdriver_home()
{
	LCDdriver_command(kLCD_RETURNHOME);	// set cursor position to zero
	_delay_us(2000);
}

//************************************************************************
void LCDdriver_clear()
{
	LCDdriver_command(kLCD_CLEARDISPLAY);	// clear display, set cursor position to zero
	_delay_us(2000);
}

//************************************************************************
void LCDdriver_setCursor(unsigned char col, unsigned char row)
{
//int row_offsets[]	=	{ 0x00, 0x40, 0x14, 0x54 };

	LCDdriver_command(kLCD_SETDDRAMADDR | (col + ((row == 0) ? 0x00 : 0x40)));
}


//***************************************************************
void	LCDdriver_Init(void)
{
uint8_t	myFunctionSetCommand;

	DDRA		=	0xff;		//*	data port as output

	myFunctionSetCommand	=		kLCD_FUNCTIONSET +	//*	function set command
									kLCD_8BITMODE +		//*	8 bit mode
									kLCD_2LINE +		//*	2 line
									kLCD_5x10DOTS;		//*	dots

//	myFunctionSetCommand	=	0x38;
	LCDdriver_command(myFunctionSetCommand);		
	_delay_us(4500);	// wait more than 4.1ms
	LCDdriver_command(myFunctionSetCommand);		
	_delay_us(200);
	LCDdriver_command(myFunctionSetCommand);		
	_delay_us(200);
	LCDdriver_command(myFunctionSetCommand);

	_delay_us(1000);

	LCDdriver_command(	kLCD_DISPLAYCONTROL +		//*	display command
						kLCD_DISPLAYON +			//*	display ON
						kLCD_CURSORON +				//*	cursor ON
						kLCD_BLINKOFF);				//*	blinking OFF

	_delay_us(200);

	LCDdriver_command(	kLCD_ENTRYMODESET +
						kLCD_ENTRYLEFT);
}


